/*
   SETENV.C:
      Modifies the master DOS environment. If the new variable and value will
      fit it deletes the existing variable then appends the new one to the
      end of the remaining environment.

      In the case of the PATH variable, it will prepend the new value in
      front of the old path. This is useful for adding a directory to
      the existing path. This can be disabled by using the command line
      parameter "/PREPEND:OFF". Default is "/PREPEND:ON".

      This is not a pretty program, but it works. It was written because I
      wanted to modify the PROMPT variable in the master environment to
      randomly generate prompt color. See SETPROMP.C.

      Written in Turbo C  version 2.0.

      Author: Brad Larsen      CIS ID#: 74504,721     DATE: 11-11-94
              Park Falls, WI
*/

#include <stdio.h>
#include <alloc.h>
#include <dos.h>
#include <string.h>
#include "proto.h"

void main(int argc, char *argv[])
{
   char      far  *env_str_address;
   char      far  *env_str_ptr;
   unsigned       max_env_size;
   unsigned       cur_env_size;
   char           name[128];
   char           new_variable[256];
   char           value[256];
   char           temp[128];
   unsigned       state = CHECK_PARAMETERS;
   unsigned       done = FALSE;
   unsigned       prepend_path = TRUE;
   int            x = 0;
   int            data_parameter = 0;


   while(!done)
   {
      switch(state)
      {
         case CHECK_PARAMETERS:
            switch(argc)
            {
               case(1):  /* no extra command arguments, display environment   */
                  get_env_info(&env_str_address, &max_env_size, &cur_env_size);
                  display_environment(env_str_address, max_env_size, cur_env_size);
                  state = FINISHED;
                  break;

               default:  /* At least one extra command line parameter */
                  for(x = 1; x < argc; x++)
                  {
                     if(!stricmp(argv[x], "/PREPEND:OFF"))
                     {
                        /* Disable path prepend function */
                        prepend_path = FALSE;
                     }
                     else if(strstr(argv[x], "="))
                     {
                        /* Store which position has the actual data */
                        data_parameter = x;
                        state = INITIALIZE;
                     }
                     else if(!stricmp(argv[x], "/?"))
                     {
                        display_help();
                        x = argc;
                        state = FINISHED;
                     }
                     else if(!stricmp(argv[x], "/HELP"))
                     {
                        display_help();
                        x = argc;
                        state = FINISHED;
                     }
                     else
                     {
                        printf("\nSyntax Error\n");
                        printf("\nType SET /? for help\n");
                        x = argc;
                        state = FINISHED;
                     }
                  }
               break;
            }
            break;

         case INITIALIZE:
            create_string(argv, new_variable, name, data_parameter);
            /* check if first argument ends in and \"=\" sign */
            /* if so delete it, as the DOS SET would          */
            if(argv[data_parameter][strlen(argv[data_parameter])-1] == '=')
            {
               get_env_info(&env_str_address, &max_env_size, &cur_env_size);
               env_str_ptr = get_name_ptr(env_str_address, strupr(argv[data_parameter]), value);
               if(env_str_ptr != NULL)
               {
                  del_variable(env_str_address, env_str_ptr, &cur_env_size);
               }
               state = FINISHED;
            }
            else
            {
               state = GET_ENV_INFO;
            }
            break;

         case GET_ENV_INFO:
             get_env_info(&env_str_address, &max_env_size, &cur_env_size);
             state = GET_VARIABLE_POINTER;
             break;

         case GET_VARIABLE_POINTER:
            env_str_ptr = get_name_ptr(env_str_address, name, value);
            /* Check if the path value should be prepended or should */
            /* the new value replace the old.                        */
            if(strstr(name, "PATH=")&& prepend_path)
            {
               if(!(new_variable[strlen(new_variable)-1] == ';'))
               {
                  strcat(new_variable, ";");
                  strcat(new_variable, value);
               }
               else
               {
                  strcat(new_variable, value);
               }
               /* Truncate path if needed.                              */
               new_variable[255] = NULL;
            }
            state = CHECK_FOR_ROOM;
            break;

         case CHECK_FOR_ROOM:
            if(env_str_ptr == NULL)   /* new variable                   */
            {
               if(check_for_room(max_env_size, cur_env_size, new_variable, 0))
               {
                  state = APPEND_VARIABLE;
               }
               else
               {
                  state = NOT_ENOUGH_ROOM;
               }
            }
            else
            {
               if(check_for_room(max_env_size, cur_env_size, new_variable,
                                 strlen(name) + strlen(value) + 1))
               {
                  state = DELETE_VARIABLE;
               }
               else
               {
                  state = NOT_ENOUGH_ROOM;
               }
            }
            break;

         case DELETE_VARIABLE:
            del_variable(env_str_address, env_str_ptr, &cur_env_size);
            state = APPEND_VARIABLE;
            break;

         case NOT_ENOUGH_ROOM:
            printf("\nNot enough room for new variable.\n");
            state = FINISHED;
            break;

         case APPEND_VARIABLE:
            append_new_variable(env_str_address, new_variable);
            state = FINISHED;
            break;

         case FINISHED:
            done = TRUE;
            break;
      }
   }
}





/*
   get_env_info:
      This routine returns a pointer to the beginning of the master DOS
      environment, the max size as set in CONFIG.SYS, and the current size.
*/
void get_env_info(char far **env_str_address, unsigned *max_env_size, unsigned *cur_env_size)
{
   unsigned  far  *env_address;
   unsigned  far  *env_size_ptr;
   unsigned  far  *env_seg_ptr;
   unsigned  far  *first_pointer;
   unsigned       first_null;
   unsigned       counter = 0;
   char      far  *ptr;

   int            vector = 0x002E * 4;      /* 0x0000:0x00B8            */
   int            offset = 0x002C;     /* offset to env segment pointer */

   first_pointer = MK_FP(0x0000, vector + 2);
   env_seg_ptr = MK_FP(*first_pointer, offset);

   /* address of start of environment */
   env_address = MK_FP(*env_seg_ptr, 0);
   ptr = (char far *) env_address;
   *env_str_address = (char far *)MK_FP(*env_seg_ptr, 0);

   /* 1 paragraph back at offset 3 is max environment size in 16 byte   */
   /* paragraphs.                                                       */
   env_size_ptr = MK_FP((*env_seg_ptr - 1), 0x0003);
   *max_env_size = 16 * (*env_size_ptr);

   /* determine current environment size. It ends with 2 NULL's.        */
   for(; !((*ptr == NULL) && (*(ptr - 1) == NULL)) ; ptr++, counter++);
   *cur_env_size = counter;
}




/*
   get_name_ptr:
      Scans the master environment for a given string and returns a
      pointer to the beginning of it. Else returns NULL. It also returns
      the value of the given string if it exist. If it does not, value is
      set to a zero length string.

      Variable "name" points to an upper case string.

      Variable "value" points to a string with the same case as in the master
      environment.
*/
char far *get_name_ptr (char far *env_str_address , char *name, char *value)
{
   char far *index;
   char far *index1;
   char far *string1;
   char far *string2;
   unsigned counter;

   /*  Set index to begining of environment */
   index = env_str_address;

   /*  While not at the end of the environment */
   while(*index)
   {
      string1 = index;
      string2 = name;
      /* Loop till char's don't match or at the end of either string   */
      while((*string1 == *string2) && (*string1) && (*string2))
      {
         string1++;
         string2++;
      }

      /* Reached the end of the string I was looking for.               */
      /* Return a pointer start of this string.                         */
      if (!*string2)    /* reached the end of name string               */
      {
         index1 = index;
         for(; *index1 != '='; index1++);
         index1++;
         /* Get the value of the existing variable                      */
         for(counter = 0 ; *index1 != NULL ;)
         {
            value[counter] = *index1;
            index1++;
            counter++;
         }
         value[counter] = NULL;
         return(index);
      }

      /*  Go to the next variable in the env.                           */
      while(*(index++));
   }

   /* String is not present, return NULL.                               */
   return(NULL);
}




/*
   check_for_room:
      Returns TRUE if there is room for new variable plus a
      terminating NULL. Else returns FALSE. Takes into account if a variable
      is to be deleted.
*/
unsigned check_for_room(unsigned max_env_size, unsigned cur_env_size, char *new_variable, unsigned size_to_delete)
{
   unsigned  ret = FALSE;

   if(max_env_size > (cur_env_size + strlen(new_variable) - size_to_delete + 1))
   {
      ret = TRUE;
   }
   return(ret);
}




/*
   del_variable:
      Locates the beginning of the string to be overwritten. Advances a
      pointer to the next string and writes the remainder of environment over
      the top of the variable to be deleted
*/
void del_variable(char far *env_str_address, char far *env_str_ptr, unsigned *cur_env_size)
{
   char     far  *index;
   unsigned      done = FALSE;

   /* Go to beginning of string to be deleted                           */
   while(!(env_str_address == env_str_ptr))
   {
      env_str_address = env_str_address + 1;
   }
   /* Mark this place                                                   */
   index = env_str_address;

   /* Move to the beginning of the next string adjusting the current    */
   /* environment size as you go.                                       */
   while(!(*env_str_address == NULL))
   {
      env_str_address++;
      *cur_env_size = *cur_env_size - 1;
   }
   /* Final adjustments                                                 */
   env_str_address++;
   *cur_env_size = *cur_env_size - 1;

   /* Overwrite variable to be deleted till at end of environment       */
   while(!done)
   {
      *index = *env_str_address;
      if((*env_str_address == NULL) && (*(env_str_address - 1) == NULL))
      {
         done = TRUE;
      }
      env_str_address = env_str_address + 1;
      index = index + 1;
   }
}




/*
   append_new_variable:
      Go to the end of the environment and add the new variable, then
      add the last 2 terminating NULL's.
*/
void append_new_variable(char far *env_str_address, char *new_variable)
{
   char       *new_variable_ptr;

   new_variable_ptr = new_variable;                 /* create a pointer */

   /* Advance to end of environment                                     */
   for(;!((*env_str_address == NULL) && (*(env_str_address - 1) == NULL));)
   {
      env_str_address++;
   }

   for(; *new_variable_ptr ;)
   {
      *env_str_address = *new_variable_ptr;
      env_str_address++;
      new_variable_ptr++;
   }
   /* Make sure to end the environment with two NULL's                  */
   *env_str_address = NULL;
   *(env_str_address + 1) = NULL;
}

/*
   DISPLAY_ENVIRONMENT:
      Called when there are no command line arguments. Just as the SET
      command would do in the same case. Displays additional environment
      information that SET does not. It add the current and maximum
      environment size to the beginning of the display. It also automatically
      pauses after each screen full of information.
*/
void display_environment(char far *env_str_address, unsigned max_env_size,
                          unsigned cur_env_size)
{
   char temp_string[256];
   unsigned counter;
   unsigned line_counter = 4;
   unsigned done;
   unsigned end_env = FALSE;

   /* Additional info */
   printf("\nCurrent environment size = %d\n", cur_env_size);
   printf("Maximum environment size = %d\n\n", max_env_size);

   while(!end_env)
   {
      counter = 0;
      done = FALSE;
      while(!done)
      {
         /* Not at end of string */
         if(*env_str_address == NULL)
         {
            done = TRUE;
         }
         temp_string[counter] = *env_str_address;
         counter++;
         env_str_address++;
      }
      /* display on screen */
      printf("%s\n", temp_string);

      /* 2 NULL's in a row signal the end of environment */
      if(*env_str_address == NULL)
      {
         end_env = TRUE;
      }

      /* auto pause after each screen */
      line_counter++;
      if(line_counter >= 22)
      {
         printf("\nPress any key to continue...\n");
         for(;!kbhit(););
         getch();
         line_counter = 0;
      }
   }
}

/*
   DISPLAY_HELP:
      Called if the first command line argument does not have an "=" in it.
*/
void display_help(void)
{
   printf("\nThis utility sets or modifies variables in ");
   printf("the master DOS environment.\n");
   printf("\nIn the case of the PATH variable, it prepends\n");
   printf("the new value to the old and truncates if needed.\n");
   printf("A trailing semi-colon is not needed on the command line.\n");
   printf("By adding the command line parameter \"/PREPEND:OFF\", ");
   printf("PATH will\n");
   printf("be treated the same as other environment variables.\n");
   printf("\nSyntax:  SETENV [/PREPEND:ON|OFF] [/?] [variable=[value]]\n");
   printf("\nExamples: SETENV windir=c:\\windows , SETENV test=\n");
   printf("\nEntering SETENV alone will display current environment\n");
   printf("\nIf value is NULL, the variable will be deleted.\n");
   printf("\nPlace SETENV.EXE in a directory in your path. Then set up a\n");
   printf("DOSKEY macro to replace the internal DOS SET command.\n");
   printf("Add the line:\n");
   printf("   DOSKEY SET=setenv $1 $2 $3 $4 $5 $6 $7 $8 $9\n");
   printf("to your AUTOEXEC.BAT file.\n");
}

/*
   CREATE_STING:
      Takes the first command line argument and converts everything before
      the "=" to upper case. Passes back this converted string as well as
      the upper case variable name followed by the "=" sign.

      Case is converted because that is what the DOS SET command does. Also
      will need to search for the upper case version of the variable name as
      all DOS environment variable names are upper case.
*/
void create_string(char *argv[], char *new_variable, char *name, int data_parameter)
{
   unsigned counter = 0;
   char     temp[128];

   sprintf(new_variable, argv[data_parameter]);
   /* Put everything before "=" in temp */
   while(argv[data_parameter][counter] != '=')
   {
   temp[counter] = argv[data_parameter][counter];
   counter++;
   }

   /* add "=" and then terminate the string */
   temp[counter] = '=';
   counter++;
   temp[counter] = '\0';
   sprintf(name, temp);

   /* convert to upper case */
   strupr(name);

   /* overwrite old variable name with upper case name */
   for(counter = 0 ; name[counter] != '=' ; counter++)
   {
      new_variable[counter] = name[counter];
   }

   /* if more than one command line argument concat them   */
/*
   for(counter = 3 ; counter <= argc ; counter++)
   {
      strcat(new_variable, " ");
      strcat(new_variable, argv[counter - 1]);
   }
*/
}
